@@ -43,9 +43,7 @@
 int unix_main(int, char *[]);
 
 /* Prototypes */
-static void parse_cmdline(LPSTR);
-static int set_cwd(void);
-static int initialize_winsock(void);
+static int load_ws2(char *);
 static void ThreadFunc(void *);
 static LRESULT CALLBACK wndProc(HWND, UINT, WPARAM, LPARAM);
 static int win_main(HINSTANCE, HINSTANCE, LPSTR, int);
@@ -56,14 +54,11 @@ static char *log_txt(void);
 static void set_visible(int);
 
 /* NT Service related function */
-static int service_initialize(void);
-static int service_install(LPTSTR);
-static int service_uninstall(void);
-static int service_start(void);
-static int service_stop(void);
+static int start_service(void);
+static int install_service(void);
+static int uninstall_service(void);
 static void WINAPI service_main(DWORD, LPTSTR *);
 static void WINAPI control_handler(DWORD);
-static void error_box(char *text);
 
 /* Global variables */
 static struct LIST {
@@ -78,6 +73,7 @@ static HMENU hpopup;
 static HWND hwnd=NULL;
 static HANDLE small_icon; /* 16x16 icon */
 
+static char service_path[MAX_PATH];
 static SERVICE_STATUS serviceStatus;
 static SERVICE_STATUS_HANDLE serviceStatusHandle=0;
 static HANDLE stopServiceEvent=0;
@@ -87,133 +83,95 @@ static jmp_buf jump_buf;
 
 static char passphrase[STRLEN];
 
-GETADDRINFO s_getaddrinfo;
-FREEADDRINFO s_freeaddrinfo;
-GETNAMEINFO s_getnameinfo;
-
-static struct {
-    char config_file[STRLEN];
-    unsigned int install:1, uninstall:1, start:1, stop:1, service:1, quiet:1;
-} cmdline;
+GETADDRINFO s_getaddrinfo=NULL;
+FREEADDRINFO s_freeaddrinfo=NULL;
+GETNAMEINFO s_getnameinfo=NULL;
 
 int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
     LPSTR lpszCmdLine, int nCmdShow) {
 
+    char exe_file_name[MAX_PATH], dir[MAX_PATH], *ptr;
+    static struct WSAData wsa_state;
+
     ghInst=hInstance;
 
-    parse_cmdline(lpszCmdLine); /* setup global cmdline structure */
-    if(set_cwd()) /* set current working directory */
-        return 1;
+    GetModuleFileName(0, exe_file_name, MAX_PATH);
 
-    if(initialize_winsock())
+    /* set current directory */
+    strcpy(dir, exe_file_name);
+    ptr=strrchr(dir, '\\'); /* last backslash */
+    if(ptr)
+        ptr[1]='\0'; /* truncate program name */
+    if(!SetCurrentDirectory(dir)) {
+        MessageBox(hwnd, "Cannot set current directory",
+            options.win32_name, MB_ICONERROR);
         return 1;
+    }
 
-    if(!setjmp(jump_buf)) { /* TRY */
-        main_initialize(cmdline.config_file, NULL);
-        if(!cmdline.service) {
-            if(cmdline.install)
-                return service_install(lpszCmdLine);
-            if(cmdline.uninstall)
-                return service_uninstall();
-            if(cmdline.start)
-                return service_start();
-            if(cmdline.stop)
-                return service_stop();
+    /* setup service_path for CreateService() */
+    strcpy(service_path, "\"");
+    strcat(service_path, exe_file_name);
+    strcat(service_path, "\" -service");
+    /* strcat(service_path, lpszCmdLine); */
+
+    if(WSAStartup(MAKEWORD( 2, 2 ), &wsa_state)) {
+        MessageBox(hwnd, "Failed to initialize winsock",
+            options.win32_name, MB_ICONERROR);
+        return 1;
+    }
+    if(!load_ws2("ws2_32.dll"))
+        load_ws2("wship6.dll");
+
+    if(!error_mode && !strcmpi(lpszCmdLine, "-service")) {
+        if(!setjmp(jump_buf))
+            main_initialize(NULL, NULL);
+        return start_service(); /* Always start service with -service option */
+    }
+
+    if(!error_mode && !setjmp(jump_buf)) { /* TRY */
+        if(!strcmpi(lpszCmdLine, "-install")) {
+            main_initialize(NULL, NULL);
+            return install_service();
+        } else if(!strcmpi(lpszCmdLine, "-uninstall")) {
+            main_initialize(NULL, NULL);
+            return uninstall_service();
+        } else { /* not -service, -install or -uninstall */
+            main_initialize(lpszCmdLine[0] ? lpszCmdLine : NULL, NULL);
         }
     }
 
     /* CATCH */
-    if(cmdline.service)
-        return service_initialize();
-    else
-        return win_main(hInstance, hPrevInstance, lpszCmdLine, nCmdShow);
+    return win_main(hInstance, hPrevInstance, lpszCmdLine, nCmdShow);
 }
 
-static void parse_cmdline(LPSTR lpszCmdLine) {
-    char line[STRLEN], *c, *opt;
-
-    safecopy(line, lpszCmdLine);
-    memset(&cmdline, 0, sizeof(cmdline));
-
-    c=line;
-    while(*c && (*c=='/' || *c=='-')) {
-        opt=c;
-        while(*c && !isspace(*c)) /* skip non-whitespaces */
-            c++;
-        while(*c && isspace(*c)) /* replace whitespaces with '\0' */
-            *c++='\0';
-        if(!strcmpi(opt+1, "install"))
-            cmdline.install=1;
-        else if(!strcmpi(opt+1, "uninstall"))
-            cmdline.uninstall=1;
-        else if(!strcmpi(opt+1, "start"))
-            cmdline.start=1;
-        else if(!strcmpi(opt+1, "stop"))
-            cmdline.stop=1;
-        else if(!strcmpi(opt+1, "service"))
-            cmdline.service=1;
-        else if(!strcmpi(opt+1, "quiet"))
-            cmdline.quiet=1;
-        else { /* option to be processed in options.c */
-            safecopy(cmdline.config_file, opt);
-            return; /* no need to parse other options */
-        }
-    }
-
-    if(*c) /* configuration file specified in the command line */
-        safecopy(cmdline.config_file, c);
-    else /* default configuration file name */
-        safecopy(cmdline.config_file, "stunnel.conf");
-}
-
-static int set_cwd(void) {
-    char *c, errmsg[STRLEN], exe_file_name[STRLEN];
-
-    GetModuleFileName(0, exe_file_name, STRLEN);
-    c=strrchr(exe_file_name, '\\'); /* last backslash */
-    if(c) /* found */
-        c[1]='\0'; /* truncate program name */
-    if(!SetCurrentDirectory(exe_file_name)) {
-        safecopy(errmsg, "Cannot set directory to ");
-        safeconcat(errmsg, exe_file_name);
-        MessageBox(hwnd, errmsg, "stunnel", MB_ICONERROR);
-        return 1;
-    }
-    return 0;
-}
-
-/* try to load winsock2 resolver functions from a specified dll name */
-static int initialize_winsock() {
-    static struct WSAData wsa_state;
+/* Try to load winsock2 resolver functions from a specified dll name */
+static int load_ws2(char *name) {
     HINSTANCE handle;
     
-    if(WSAStartup(MAKEWORD( 2, 2 ), &wsa_state)) {
-        MessageBox(hwnd, "Failed to initialize winsock",
-            "stunnel", MB_ICONERROR);
-        return 1; /* error */
-    }
-    handle=LoadLibrary("ws2_32.dll"); /* IPv6 in Windows XP or higher */
-    if(handle) {
-        s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
-        s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
-        s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
-        if(s_getaddrinfo && s_freeaddrinfo && s_getnameinfo)
-            return 0; /* IPv6 detected -> OK */
+    handle=LoadLibrary(name);
+    if(!handle)
+        return 0;
+
+    s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
+    if(!s_getaddrinfo) {
         FreeLibrary(handle);
+        return 0;
     }
-    handle=LoadLibrary("wship6.dll"); /* Experimental IPv6 for Windows 2000 */
-    if(handle) {
-        s_getaddrinfo=(GETADDRINFO)GetProcAddress(handle, "getaddrinfo");
-        s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
-        s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
-        if(s_getaddrinfo && s_freeaddrinfo && s_getnameinfo)
-            return 0; /* IPv6 detected -> OK */
+
+    s_freeaddrinfo=(FREEADDRINFO)GetProcAddress(handle, "freeaddrinfo");
+    if(!s_freeaddrinfo) {
         FreeLibrary(handle);
+        return 0;
     }
-    s_getaddrinfo=NULL;
-    s_freeaddrinfo=NULL;
-    s_getnameinfo=NULL;
-    return 0; /* IPv4 detected -> OK */
+
+    s_getnameinfo=(GETNAMEINFO)GetProcAddress(handle, "getnameinfo");
+    if(!s_getnameinfo) {
+        FreeLibrary(handle);
+        return 0;
+    }
+
+    FreeLibrary(handle);
+    return 1;
 }
 
 static int win_main(HINSTANCE hInstance, HINSTANCE hPrevInstance,
@@ -547,7 +505,7 @@ static void save_file(HWND hwnd) {
     if((hFile=CreateFile((LPCSTR)szFileName, GENERIC_WRITE,
             0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL,
             (HANDLE) NULL))==INVALID_HANDLE_VALUE) {
-        error_box("CreateFile");
+        MessageBox(hwnd, "File open failed", options.win32_name, MB_ICONERROR);
         return;
     }
 
@@ -556,7 +514,7 @@ static void save_file(HWND hwnd) {
     bResult=WriteFile(hFile, txt, nToWrite, &nWritten, NULL);
     free(txt);
     if(!bResult)
-        error_box("WriteFile");
+        MessageBox(hwnd, "File write failed", options.win32_name, MB_ICONERROR);
     CloseHandle(hFile);
 }
 
@@ -644,70 +602,68 @@ void exit_stunnel(int code) { /* used instead of exit() on Win32 */
     longjmp(jump_buf, 1);
 }
 
-static int service_initialize(void) {
+static int start_service(void) {
     SERVICE_TABLE_ENTRY serviceTable[]={
         {options.win32_service, service_main},
         {0, 0}
     };
 
     if(!StartServiceCtrlDispatcher(serviceTable)) {
-        error_box("StartServiceCtrlDispatcher");
+        MessageBox(hwnd, "Unable to start the service",
+            options.win32_name, MB_ICONERROR);
         return 1;
     }
     return 0; /* NT service started */
 }
 
-static int service_install(LPSTR lpszCmdLine) {
+static int install_service(void) {
     SC_HANDLE scm, service;
-    char exe_file_name[STRLEN], service_path[STRLEN];
 
     scm=OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
     if(!scm) {
-        error_box("OpenSCManager");
+        MessageBox(hwnd, "Failed to open service control manager",
+            options.win32_name, MB_ICONERROR);
         return 1;
     }
-    GetModuleFileName(0, exe_file_name, STRLEN);
-    safecopy(service_path, "\"");
-    safeconcat(service_path, exe_file_name);
-    safeconcat(service_path, "\" -service ");
-    safeconcat(service_path, lpszCmdLine);
     service=CreateService(scm,
         options.win32_service, options.win32_service, SERVICE_ALL_ACCESS,
         SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS,
         SERVICE_AUTO_START, SERVICE_ERROR_NORMAL, service_path,
         NULL, NULL, NULL, NULL, NULL);
     if(!service) {
-        error_box("CreateService");
+        MessageBox(hwnd, "Failed to create a new service",
+            options.win32_name, MB_ICONERROR);
         CloseServiceHandle(scm);
         return 1;
     }
-    if(!cmdline.quiet)
-        MessageBox(hwnd, "Service installed",
-            options.win32_name, MB_ICONINFORMATION);
+    MessageBox(hwnd, "Service installed", options.win32_name,
+        MB_ICONINFORMATION);
     CloseServiceHandle(service);
     CloseServiceHandle(scm);
     return 0;
 }
 
-static int service_uninstall(void) {
+static int uninstall_service(void) {
     SC_HANDLE scm, service;
     SERVICE_STATUS serviceStatus;
 
     scm=OpenSCManager(0, 0, SC_MANAGER_CONNECT);
     if(!scm) {
-        error_box("OpenSCManager");
+        MessageBox(hwnd, "Failed to open service control manager",
+            options.win32_name, MB_ICONERROR);
         return 1;
     }
     service=OpenService(scm, options.win32_service,
         SERVICE_QUERY_STATUS | DELETE);
     if(!service) {
-        if(!cmdline.quiet)
-            error_box("OpenService");
+        MessageBox(hwnd, "Failed to open the service",
+            options.win32_name, MB_ICONERROR);
         CloseServiceHandle(scm);
         return 1;
     }
     if(!QueryServiceStatus(service, &serviceStatus)) {
-        error_box("QueryServiceStatus");
+        MessageBox(hwnd, "Failed to query service status",
+            options.win32_name, MB_ICONERROR);
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
@@ -720,114 +676,14 @@ static int service_uninstall(void) {
         return 1;
     }
     if(!DeleteService(service)) {
-        error_box("DeleteService");
-        CloseServiceHandle(service);
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    if(!cmdline.quiet)
-        MessageBox(hwnd, "Service uninstalled", options.win32_name,
-            MB_ICONINFORMATION);
-    CloseServiceHandle(service);
-    CloseServiceHandle(scm);
-    return 0;
-}
-
-static int service_start(void) {
-    SC_HANDLE scm, service;
-    SERVICE_STATUS serviceStatus;
-
-    scm=OpenSCManager(0, 0, SC_MANAGER_CONNECT);
-    if(!scm) {
-        error_box("OpenSCManager");
-        return 1;
-    }
-    service=OpenService(scm, options.win32_service,
-        SERVICE_QUERY_STATUS | SERVICE_START);
-    if(!service) {
-        error_box("OpenService");
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    if(!StartService(service, 0, NULL)) {
-        error_box("StartService");
-        CloseServiceHandle(service);
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    do {
-        sleep(1);
-        if(!QueryServiceStatus(service, &serviceStatus)) {
-            error_box("QueryServiceStatus");
-            CloseServiceHandle(service);
-            CloseServiceHandle(scm);
-            return 1;
-        }
-    } while(serviceStatus.dwCurrentState==SERVICE_START_PENDING);
-    if(serviceStatus.dwCurrentState!=SERVICE_RUNNING) {
-        MessageBox(hwnd, "Failed to start service",
+        MessageBox(hwnd, "Failed to delete the service",
             options.win32_name, MB_ICONERROR);
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
-    if(!cmdline.quiet)
-        MessageBox(hwnd, "Service started", options.win32_name,
-            MB_ICONINFORMATION);
-    CloseServiceHandle(service);
-    CloseServiceHandle(scm);
-    return 0;
-}
-
-static int service_stop(void) {
-    SC_HANDLE scm, service;
-    SERVICE_STATUS serviceStatus;
-
-    scm=OpenSCManager(0, 0, SC_MANAGER_CONNECT);
-    if(!scm) {
-        error_box("OpenSCManager");
-        return 1;
-    }
-    service=OpenService(scm, options.win32_service,
-        SERVICE_QUERY_STATUS | SERVICE_STOP);
-    if(!service) {
-        if(!cmdline.quiet)
-            error_box("OpenService");
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    if(!QueryServiceStatus(service, &serviceStatus)) {
-        error_box("QueryServiceStatus");
-        CloseServiceHandle(service);
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    if(serviceStatus.dwCurrentState==SERVICE_STOPPED) {
-        if(!cmdline.quiet)
-            MessageBox(hwnd, "The service is already stopped",
-                options.win32_name, MB_ICONERROR);
-        CloseServiceHandle(service);
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    if(!ControlService(service, SERVICE_CONTROL_STOP, &serviceStatus)) {
-        error_box("ControlService");
-        CloseServiceHandle(service);
-        CloseServiceHandle(scm);
-        return 1;
-    }
-    do {
-        sleep(1);
-        if(!QueryServiceStatus(service, &serviceStatus)) {
-            error_box("QueryServiceStatus");
-            CloseServiceHandle(service);
-            CloseServiceHandle(scm);
-            return 1;
-        }
-    } while(serviceStatus.dwCurrentState!=SERVICE_STOPPED);
-    if(!cmdline.quiet)
-        MessageBox(hwnd, "Service stopped", options.win32_name,
-            MB_ICONINFORMATION);
+    MessageBox(hwnd, "Service uninstalled", options.win32_name,
+        MB_ICONINFORMATION);
     CloseServiceHandle(service);
     CloseServiceHandle(scm);
     return 0;
@@ -907,18 +763,4 @@ static void WINAPI control_handler(DWORD controlCode) {
     SetServiceStatus(serviceStatusHandle, &serviceStatus);
 }
 
-static void error_box(char *text) {
-    char to_print[STRLEN]; 
-    char *buff;
-    long dw;
-    
-    dw=GetLastError(); 
-    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
-        NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-        (LPTSTR) &buff, 0, NULL);
-    snprintf(to_print, STRLEN, "%s: error %ld: %s", text, dw, buff); 
-    MessageBox(hwnd, to_print, options.win32_name, MB_ICONERROR);
-    LocalFree(buff);
-}
-
 /* End of gui.c */
